/* Copyright (c) 2012 MJ Rutter 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street,
 * Fifth Floor, Boston, MA  02110-1301, USA.
 */ 

/* CCITT Group 4 2D encode an image
 * or CCITT Group 3 1D encode an image
 *
 * NB 1 is white and 0 is black
 *
 * MJR 7/12
 */

#include<stdio.h>
#include<stdlib.h>

#include"bmp2eps.h"

#define MAD 0     /* For copious G4 debugging */
#define G3EOL 0   /* For EOL codes in G3 output */

void sbits(unsigned char *c,int st, int len, int data); /* Found in to_lzw2.c */
static unsigned next_change(unsigned char *img, int st, int end, char col);
static void g3code(unsigned char *buff, unsigned int *off, int repeat, int col);

unsigned to_group4(struct bitmap *image, unsigned char **rtn){
  unsigned char col, *p, *out;
  unsigned int j,off,repeat,debug;
  unsigned int a0,a1,a2,b1,b2;
  unsigned char *ref,*code;

  int vdata[7]={2,2,2,1,3,3,3};
  int vlen[7]={7,6,3,1,3,6,7};
  int *vert,*vert_len;

  vert=vdata+3;      /* vert is an array indexed from -3 to 3 */
  vert_len=vlen+3;   /* ditto */
  debug=image->flags&DEBUG_MASK;

  if (debug>2) fprintf(stderr,"CCITT4 called\n");

  out=malloc(image->data_len+10+image->width/1000);
  if (!out){
    fprintf(stderr,"Malloc error in group4 conversion\n");
    exit(1);
  }

  off=0;
  code=image->data;
  ref=NULL;

  for(j=0;j<image->height;j++){
    col=1;
    code=image->data+j*((image->width+7)>>3);

    a0=0;
    a1=next_change(code,0,image->width,1);
    a2=next_change(code,a1,image->width,0);

    if (j==0)
      b1=b2=image->width+1; /* Initial reference line of all white */
    else{
      ref=image->data+(j-1)*((image->width+7)>>3);
      b1=next_change(ref,0,image->width,1);
      b2=next_change(ref,b1,image->width,0);
    }

    do{
#if MAD
      fprintf(stderr,"a0=%d a1=%d a2=%d b1=%d b2=%d\n",a0,a1,a2,b1,b2);
#endif

      if (b2<a1){ /* Pass mode */
#if MAD
	fprintf(stderr,"Pass\n");
#endif
	sbits(out+(off>>3),off&7,4,1);
	off+=4;
	a0=b2;
        b1=next_change(ref,a0,image->width,col);
	b2=next_change(ref,b1,image->width,col^1);
      }
      else if (((a1-b1)<=3)&&((b1-a1)<=3)){ /* Vertical mode */
#if MAD
	fprintf(stderr,"Vertical %d\n",b1-a1);
#endif
	sbits(out+(off>>3),off&7,vert_len[b1-a1],vert[b1-a1]);
        off+=vert_len[b1-a1];
        col^=1;
        a0=a1;
        a1=a2;
	a2=next_change(code,a1,image->width,col^1);
	if (b1<=a1){
	  b1=next_change(ref,a0,image->width,col^1);
	  b1=next_change(ref,b1,image->width,col);
	  b2=next_change(ref,b1,image->width,col^1);
	}
      }
      else{ /* Horizontal mode */
#if MAD
	fprintf(stderr,"Horizontal\n");
#endif
	sbits(out+(off>>3),off&7,3,1);
	off+=3;
	repeat=a1-a0;
#if MAD
        fprintf(stderr,"col %d repeat %d code %x length %d\n",col,repeat,
		term[col][repeat],term_len[col][repeat]);
#endif
        g3code(out,&off,repeat,(int)col);
        col^=1;
	repeat=a2-a1;
#if MAD
        fprintf(stderr,"col %d repeat %d code %x length %d\n",col,repeat,
		term[col][repeat],term_len[col][repeat]);
#endif
        g3code(out,&off,repeat,(int)col);
        col^=1;
	a0=a2;
        a1=next_change(code,a0,image->width,col);
	a2=next_change(code,a1,image->width,col^1);
	if (b1<=a1){
	  b1=next_change(ref,a0,image->width,col^1);
	  b1=next_change(ref,b1,image->width,col);
	  b2=next_change(ref,b1,image->width,col^1);
	}
      }
      if ((off>>3)>image->data_len){  /* Compression useless */
	free(out);
	rtn=NULL;
	if (debug>1) fprintf(stderr,"CCITT4 aborted on line %d due to size\n",j);
	return(0);
      }
    } while(a0<image->width);
  }

  p=realloc(out,(off+7)>>3);
  if (p) out=p;
  if (debug>2) fprintf(stderr,"CCITT4 exiting with %d bytes\n",(off+7)>>3);
  *rtn=out;
  return((off+7)>>3);
}

unsigned to_group3(struct bitmap *image, unsigned char **rtn){
  unsigned char col, *p, *out;
  unsigned int i,j,off,ioff,repeat,debug,line_end;
#if G3EOL
  unsigned int eol=1,eol_len=12;
#endif

  debug=image->flags&DEBUG_MASK;

  if (debug>2) fprintf(stderr,"CCITT called\n");

  out=malloc(image->data_len+5+image->width/1000);
  if (!out){
    fprintf(stderr,"Malloc error in group3 conversion\n");
    exit(1);
  }

  off=ioff=0;
#if G3EOL
  sbits(out,0,eol_len,eol);
  off=eol_len;
#endif

  for(j=0;j<image->height;j++){
    col=1;
    line_end=ioff+image->width;
    while(ioff<line_end){
      i=next_change(image->data,ioff,line_end,col);
      repeat=i-ioff;
      ioff=i;
      g3code(out,&off,repeat,col);
      col^=1;
      if ((off>>3)>image->data_len){  /* Compression useless */
        free(out);
        rtn=NULL;
        if (debug>1) fprintf(stderr,"CCITT aborted on line %d\n",j);
        return(0);
      }
    }
#if G3EOL
    sbits(out+(off>>3),off&7,eol_len,eol); /* EOL */
    off+=eol_len;
#endif
    if (ioff&7) ioff+=8-(ioff&7);    /* Lines start on byte boundaries */
   }
  /* EOD */
  p=realloc(out,(off+7)>>3);
  if (p) out=p;
  if (debug>2) fprintf(stderr,"CCITT exiting with %d bytes\n",(off+7)>>3);
  *rtn=out;
  return((off+7)>>3);
}

/* Find next bit in bitmap which is different to col. All offsets etc as bits */

static unsigned next_change(unsigned char *img, int st, int end, char col){
  int go;
  char col8;

  if (st>=end) return end;
	
#define byte (st>>3)
#define bit (st&7)

  go=1;
  while (bit){
    if ((st<end)&&
	(((img[byte]>>(7-bit))&1)==col)) st++;
    else {go=0;break;}
  }

  if (go==0) return st;

  col8=col*255;
  while ((st<end-7)&&(img[byte]==col8)) st+=8;

  while (bit<=7){
    if ((st<end)&&
	(((img[byte]>>(7-bit))&1)==col)) st++;
    else break;
  }

  return st; 
#undef byte
#undef bit
}

static void g3code(unsigned char *out, unsigned int *off, int repeat, int col){

#include "ccitt_huff.h"

  while (repeat>2560){
    sbits(out+(*off>>3),*off&7,12,0x1f);
    *off+=12;
    repeat-=2560;
  }
  if (repeat>=64){
    sbits(out+(*off>>3),*off&7,makeup_len[col][(repeat>>6)-1],
	  makeup[col][(repeat>>6)-1]);
    *off+=makeup_len[col][(repeat>>6)-1];
  }
  sbits(out+(*off>>3),*off&7,term_len[col][repeat&63],
	term[col][repeat&63]);
  *off+=term_len[col][repeat&63];
}
