
/* Copyright (c) 2015 MJ Rutter 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street,
 * Fifth Floor, Boston, MA  02110-1301, USA.
 */ 

#include<stdio.h>
#include <stdlib.h>  /* malloc */
#include "bmp2eps.h"

void palette_check(struct bitmap *image);

void colour_count(struct bitmap *image);

void stripalpha(struct bitmap *image);

/* Remove unnecessary alpha channel from RGBA data */
/* Alpha=0 means transparent */

void dealpha(struct bitmap *image){
  unsigned int i,j,size,rgb;
  int m,debug;
  unsigned char *p,*d,c1,c2,c3;
  static unsigned char buff[3];

  debug=image->flags&DEBUG_MASK;

  if (image->depth/image->components!=8) return;

  size=image->width*image->height;

  m=1;
  p=image->data+image->components-1;
  for(i=0;i<size;i++){
    if (((*p)!=0)&&((*p)!=255)){
      m=0;
      break;
    }
    p+=image->components;
  }

  if(m==0){
    if (debug>1) fprintf(stderr,"Alpha channel not simple mask\n");
    /* Now check if all transparent colours are the same */
    m=0;
    p=image->data+image->components-1;
    for(i=0;i<size;i++){
      if((*p)==0) break;
      p+=image->components;
    }
    if (image->components==4){
      rgb=((*(p-3))<<16)+((*(p-2))<<8)+*(p-1);
      for(;i<size;i++){
	if ((*p)==0){
	  if(((*(p-3))<<16)+((*(p-2))<<8)+*(p-1)!=rgb){
	    m=1;
	    break;
	  }
	}
	p+=image->components;
      }
      if (m){
	if (debug) fprintf(stderr,"Entropy in fully transparent pixels."
			   " Removing.\n");
	if (image->bkgd){
	  c1=image->bkgd[1];
	  c2=image->bkgd[2];
	  c3=image->bkgd[3];
	}
	else{
	  c1=c2=c3=0; /* Best compression as a=0 too */
	}
	p=image->data+image->components-1;
	for(i=0;i<size;i++){
	  if (*p==0){
	    *(p-3)=c1;
	    *(p-2)=c2;
	    *(p-1)=c3;
	  }
	  p+=image->components;
	}
      }
    }
    else if (image->components==2){
      rgb=*(p-1);
      for(;i<size;i++){
	if ((*p)==0){
	  if(*(p-1)!=rgb){
	    m=1;
	    break;
	  }
	}
	p+=image->components;
      }
      if (m){
	if (debug) fprintf(stderr,"Entropy in fully transparent pixels."
			   " Removing.\n");
	if (image->bkgd){
	  c1=image->bkgd[1];
	}
	else{
	  c1=0xff;
	}
	p=image->data+image->components-1;
	for(i=0;i<size;i++){
	  if (*p==0){
	    *(p-1)=c1;
	  }
	  p+=image->components;
	}
      }
    }
    return;
  }

  if (debug>1) fprintf(stderr,"Alpha channel is simple mask\n");
  
  m=1;
  p=image->data+image->components-1;
  for(i=0;i<size;i++){
    if ((*p)==0){
      m=0;
      break;
    }
    p+=image->components;
  }

  if (m!=0){
    if (debug)
      fprintf(stderr,"Alpha channel marks all pixels as fully opaque!\n");
    stripalpha(image);
    return;
  }

  if (image->components==4){
    p-=3;
    rgb=((*p)<<16)+((*(p+1))<<8)+*(p+2);

    if (debug>2) fprintf(stderr,"First transparent colour is %6x\n",rgb);

    m=1;
    p=image->data;
    for(i=0;i<size;i++){
      p+=3; /* skip rgb */
      if ((*p)==0){
	if(rgb!=((*(p-3))<<16)+((*(p-2))<<8)+*(p-1)){
	  m=0;
	  break;
	}
      }
      p++;
    }

    if (m==1){
      if (debug>1)
	fprintf(stderr,"All transparent pixels are the same colour\n");
    }
    else{
      if (debug>1){
	fprintf(stderr,"Not all transparent pixels are the same colour\n");
	if (debug>2) fprintf(stderr,"Found %2x%2x%2x\n",*(p-3),*(p-2),*(p-1));
      }
      return;
    }

    p=image->data;
    for(i=0;i<size;i++){
      if(rgb==((*p)<<16)+((*(p+1))<<8)+*(p+2)){
	if(*(p+3)!=0){
	  m=0;
	  break;
	}
      }
      p+=4;
    }

    if (m==1){
      if (debug>1)
	fprintf(stderr,"All pixels of that colour are transparent\n");
    }
    else{
      if (debug>1)
	fprintf(stderr,"Not all pixels of that colour are transparent\n");
      /* But maybe we can reduce to paletted with a duplicate entry */
      if ((image->flags&KEEP24)==0){
	colour_count(image);
	if (image->type==PALETTE){
	  if (debug)
	    fprintf(stderr,"Reduced to paletted with duplicate entry\n");
	  return;
	}
      }
      return;
    }  
    image->trans=buff;
    buff[0]=rgb>>16;
    buff[1]=(rgb>>8)&0xff;
    buff[2]=rgb&0xff;

    p=d=image->data;
    for(i=0;i<image->height;i++){
      for(j=0;j<image->width;j++){
	*d++=*p++;
	*d++=*p++;
	*d++=*p++;
	p++;
      }
    }

    image->depth=24;
    image->data_len=(image->width*image->depth+7)/8*image->height;
    image->colour=RGB;
    image->components=3;

    if (debug)
      fprintf(stderr,"Alpha channel reduced to single transparent colour\n");
  }
  else if (image->components==2){
    p--;
    rgb=*p;

    if (debug>2) fprintf(stderr,"First transparent colour is %2x\n",rgb);

    m=1;
    p=image->data+1;
    for(i=0;i<size;i++){
      if ((*p)==0){
	if(rgb!=*(p-1)){
	  m=0;
	  break;
	}
      }
      p+=2;
    }

    if (m==1){
      if (debug>1)
	fprintf(stderr,"All transparent pixels are the same colour\n");
    }
    else{
      if (debug>1){
	fprintf(stderr,"Not all transparent pixels are the same colour\n");
	if (debug>2) fprintf(stderr,"Found %2x\n",*(p-1));
      }
      return;
    }

    p=image->data;
    for(i=0;i<size;i++){
      if(rgb==(*p)){
	if(*(p+1)!=0){
	  m=0;
	  break;
	}
      }
      p+=2;
    }

    if (m==1){
      if (debug>1)
	fprintf(stderr,"All pixels of that colour are transparent\n");
    }
    else{
      if (debug>1)
	fprintf(stderr,"Not all pixels of that colour are transparent\n");
      /* But maybe we can reduce to paletted with a duplicate entry */
      if ((image->flags&KEEP24)==0){
	colour_count(image);
	if (image->type==PALETTE){
	  if (debug)
	    fprintf(stderr,"Reduced to paletted with duplicate entry\n");
	  return;
	}
      }
      return;
    }  
    image->trans=buff;
    buff[0]=rgb;

    p=d=image->data;
    for(i=0;i<image->height;i++){
      for(j=0;j<image->width;j++){
	*d++=*p++;
	p++;
      }
    }

    image->depth=8;
    image->data_len=(image->width*image->depth+7)/8*image->height;
    image->colour=GREY;
    image->components=1;

    fprintf(stderr,"Alpha channel reduced to single transparent colour\n");

  }
}


/* Convert RGBA to A1RGB where A is 0 or 255 after conversion */
void rgba2argb(struct bitmap *image){
  unsigned int i,j,debug,bpp;
  unsigned char c,*p;

  debug=image->flags&DEBUG_MASK;
  if (debug) fprintf(stderr,"Converting alpha to single bit\n");

  if ((image->depth/image->components)!=8){
    fprintf(stderr,"Invalid call to rgba2argb\n");
    exit(1);
  }

  bpp=image->components;
  p=image->data;
  for(i=0;i<image->width*image->height;i++){
    c=*(p+bpp-1);
    for(j=bpp-1;j>0;j--)
      *(p+j)=*(p+j-1);
    if (c&0x80) *p=255;
    else *p=0;
    p+=bpp;
  }

  if (image->colour==RGBA) image->colour=A1RGB;
  if (image->colour==GREYA) image->colour=A1GREY;

}

/* Remove alpha channel */

void stripalpha(struct bitmap *image){
  unsigned int debug,i,j,alpha;
  unsigned char *p,*d,*dat,*back;

  debug=image->flags&DEBUG_MASK;
  dat=image->data;
  back=image->bkgd;

  if (debug) fprintf(stderr,"Stripping alpha channel\n");

  if ((image->depth!=16)&&(image->depth!=32)){
    fprintf(stderr,"Alpha not supported at depth %d.\n",image->depth);
    exit(1);
  }

  if(!back){
    if (debug) fprintf(stderr,
		       "Warning: stripping alpha with no "
		       "background colour specified.\n");
    if (image->depth==16){
      p=d=dat;
      for(i=0;i<image->height;i++){
	for(j=0;j<image->width;j++){
	  *d++=*p++;
	  p++;
	}
      }
      image->depth=8;
      image->colour=GREY;
      image->components=1;
    }
    else if (image->depth==32){
      p=d=dat;
      for(i=0;i<image->height;i++){
	for(j=0;j<image->width;j++){
	  *d++=*p++;
	  *d++=*p++;
	  *d++=*p++;
	  p++;
	}
      }
      image->depth=24;
      image->colour=RGB;
      image->components=3;
    }
    image->data_len=(image->width*image->depth+7)/8*image->height;
    return;
  }

  if (debug>1){
    fprintf(stderr,"Background colour is ");
    if (image->depth==16) fprintf(stderr,"%2x\n",back[0]);
    else fprintf(stderr,"%2x %2x %2x\n",back[0],back[1],back[2]);
  }
  if (image->depth==16){
    p=d=dat;
    for(i=0;i<image->height;i++){
      for(j=0;j<image->width;j++){
	alpha=*(p+1);
	if (alpha==255)
	  *d++=*p++;
	else if (alpha==0){
	  *d++=back[0];
	  p++;
	}
	else
	  *d++=(alpha*(*p++)+(255-alpha)*back[0])/255;
	p++;
      }
    }
    image->depth=8;
    image->colour=GREY;
    image->components=1;
  }
  else if (image->depth==32){
    p=d=dat;
    for(i=0;i<image->height;i++){
      for(j=0;j<image->width;j++){
	alpha=*(p+3);
	if (alpha==255){
	  *d++=*p++;
	  *d++=*p++;
	  *d++=*p++;
	}
	else if (alpha==0){
	  *d++=back[0];
	  *d++=back[1];
	  *d++=back[2];
	  p+=3;
	}
	else{
	  *d++=(alpha*(*p++)+(255-alpha)*back[0])/255;
	  *d++=(alpha*(*p++)+(255-alpha)*back[1])/255;
	  *d++=(alpha*(*p++)+(255-alpha)*back[2])/255;
	}
	p++;
      }
    }
    image->depth=24;
    image->colour=RGB;
    image->components=3;
  }
  image->data_len=(image->width*image->depth+7)/8*image->height;
}

void stripalphapal(struct bitmap *image){
  unsigned int i,debug,alpha;
  unsigned char bcol[3],*p;
  
  if (image->colour!=PALETTE){
    fprintf(stderr,"stripalphapal() called on non-paletted image!\n");
    return;
  }

  if (!image->apalette){
    fprintf(stderr,"stripalphapal() called with no alpha present!\n");
    return;
  }

  debug=image->flags&DEBUG_MASK;

  if(!image->bkgd){
    if (debug) fprintf(stderr,
		       "Warning: stripping alpha with no "
		       "background colour specified.\n");
    free(image->apalette);
    image->apalette=NULL;
    image->apal_len=0;
    return;
  }

  if(image->bkgd[0]==1){
    i=image->bkgd[1];
    bcol[0]=image->palette[3*i];
    bcol[1]=image->palette[3*i+1];
    bcol[2]=image->palette[3*i+2];
  }
  else{
    bcol[0]=image->bkgd[1];
    bcol[1]=image->bkgd[2];
    bcol[2]=image->bkgd[3];
  }
  if (debug>1){
    fprintf(stderr,"Background colour is %2x %2x %2x\n",
	    bcol[0],bcol[1],bcol[2]);
  }

  p=image->palette;
  for(i=0;i<image->apal_len;i++){
    alpha=image->apalette[i];
    *p=(alpha*(*p)+(255-alpha)*bcol[0])/255;
    p++;
    *p=(alpha*(*p)+(255-alpha)*bcol[1])/255;
    p++;
    *p=(alpha*(*p)+(255-alpha)*bcol[2])/255;
    p++;
  }

  free(image->apalette);
  image->apalette=NULL;
  image->apal_len=0;

}

void rgba2smask(struct bitmap *image){
  unsigned int debug,i,j,c,size;
  unsigned char *s,*d,*m;

  debug=image->flags&DEBUG_MASK;

  if ((image->depth/image->components)!=8){
    fprintf(stderr,"Invalid call to rgba2smask\n");
    exit(1);
  }

  image->smask=malloc(sizeof(struct bitmap));
  if (!image->smask) {
    fprintf(stderr,"Malloc error for struct in rgba2smask\n");
    exit(1);
  }
  image->smask->height=image->height;
  image->smask->width=image->width;
  image->smask->data_len=image->height*image->width;
  image->smask->depth=image->depth/image->components;
  image->smask->components=1;
  image->smask->flags=image->flags;
  image->smask->colour=GREY;
  image->smask->type=BMP;
  image->smask->compression=image->compression;
  image->smask->compress_level=image->compress_level;
  image->smask->smask=NULL;
  image->smask->trans=NULL;
  image->smask->bkgd=NULL;
  image->smask->data=malloc(image->smask->data_len);
  if (!image->smask->data) {
    fprintf(stderr,"Malloc error for data in rgba2smask\n");
    exit(1);
  }

  if((image->apalette)&&(image->compression!=PNG)&&(image->type==PALETTE)
     &&(image->depth==8)){
    if (debug>1) fprintf(stderr,
			 "Constructing smask from transparency in palette\n");
    s=image->data;
    d=image->smask->data;
    for(i=0;i<image->height;i++){
      for(j=0;j<image->width;j++){
	c=*(s++);
	if (c>=image->apal_len) *d=255;
	else *d=image->apalette[c];
	d++;
      }
    }
    free(image->apalette);
    image->apalette=NULL;
    image->apal_len=0;
    return;
  }
  
  s=d=image->data;
  m=image->smask->data;
  c=image->components-1;

  size=image->width*image->height;
  for(i=0;i<size;i++){
    for(j=0;j<c;j++)
      *(d++)=*(s++);
    *(m++)=*(s++);
  }

  if (image->colour==RGBA) image->colour=RGB;
  if (image->colour==GREYA) image->colour=GREY;
  image->components--;
  image->depth=image->components*8;
  image->data_len=image->height*image->width*image->components;
}

/* CHange from palette with alpha to single transparent color */
void alphapal2tcol(struct bitmap *image){
  int i,tind,size;
  int j,j1,j2;
  int debug;

  debug=image->flags&DEBUG_MASK;

  if (image->colour!=PALETTE){
    fprintf(stderr,"alphapal2tcol() called on non-paletted image!\n");
    return;
  }

  if (!image->apalette) return;
  if ((image->depth!=8)&&(image->depth!=4)) return;
  
  /* Find transparent entry in palette */

  for(i=0;i<image->apal_len;i++)
    if (image->apalette[i]<128) break;

  if (image->apalette[i]>=128){ /* No entry was more than 50% transparent */
    if (debug) fprintf(stderr,"Discarding alpha palette"
		       " -- all entries >50%% opaque\n");
    free(image->apalette);
    image->apalette=NULL;
    image->apal_len=0;
    return;
  }

  if (debug) fprintf(stderr,"Replacing alpha palette"
		     " with single transparent entry\n");
  tind=i;
  if (image->bkgd){
    if (image->bkgd[0]==1)
      for(j=0;j<3;j++)
	image->palette[tind+j]=image->palette[image->bkgd[1]+j];
    else if (image->bkgd[0]==3)
      for(j=0;j<3;j++)
	image->palette[tind+j]=image->bkgd[1+j];
  }

  if (image->depth==8){
    size=image->width*image->height;
    for(i=0;i<size;i++){
      j=image->data[i];
      if ((j<image->apal_len)&&(image->apalette[j]<128))
	image->data[i]=tind;
    }
  }
  else if (image->depth==4){
    size=image->height*((image->width+1)/2);
    for(i=0;i<size;i++){
      j=image->data[i];
      j1=j&0xf;
      if ((j1<image->apal_len)&&(image->apalette[j1]<128))
	j1=tind;
      j2=(j>>4)&0xf;
      if ((j2<image->apal_len)&&(image->apalette[j2]<128))
	j2=tind;
      j=j1+(j2<<4);
      image->data[i]=j;
    }
  }
  

  free(image->apalette);
  image->apalette=NULL;
  image->apal_len=0;

  image->trans=malloc(1);
  if (!image->trans) error_exit ("Malloc error for one byte!");
  *image->trans=tind;
  
  palette_check(image);
  
}
