/* Copyright (c) 2010 MJ Rutter 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street,
 * Fifth Floor, Boston, MA  02110-1301, USA.
 */ 

/* de MS RLE encode (4 or 8 bit BMP version) an image
 * MJR 6/04
 */

#include <stdio.h>
#include <stdlib.h>

#include "bmp2eps.h"

void de_msrle(struct bitmap *image, unsigned char *in, unsigned len){

  unsigned char *out,*c_in,*c_out,code,*end,odd,even;
  unsigned i,j,delta,size,height,width;
  int eol,rle;

  height=image->height;
  width=image->width;
  rle=image->depth;
  if ((rle!=4)&&(rle!=8)){
    fprintf(stderr,"Impossible depth in de_msrle\n");
    exit(1);
  }
  image->depth=8;
  size=height*width;
  image->data_len=size;
  out=malloc(size);
  if (!out) {
    fprintf(stderr,"Malloc error in RLE decode\n");
    exit(1);
  }

  if (image->flags>1)
    fprintf(stderr,"RLE%d compression\n",rle);


  c_in=in;
  c_out=out;

  for(i=1;i<=height;i++){
    c_out=out+(height-i)*width;
    end=c_out+width-1;
    eol=0;
    while((c_out<=end)&&(eol==0)){
      code=*c_in++;
      if (code){ /* a run */
	if (rle==8) odd=even=*c_in++;
	else{
	  odd=*c_in&0xf;
	  even=(*c_in&0xf0)>>4;
	  c_in++;
	}
	for(j=0;(j<code/2)&&(c_out<=end);j++) {*c_out++=even;*c_out++=odd;}
	if (code&1) *c_out++=even;
      }else{
	code=*c_in++;
	switch (code) {
	case(0): /* EOL */
	  eol=1;
	  break;
	case(1): /* EOD */
	  eol=2;
	  break;
	case(2): /* Delta */
	  delta=*c_in++;    /* dx */
	  for(j=0;(j<delta)&&(c_out<=end);j++) {*c_out++=0;}
	  delta=*c_in++;    /* dy */
	  if (i+delta>height){
	    fprintf(stderr,"Image overflow in RLE decode\n");
	    exit(1);
	  }
	  for(j=0;j<delta*width;j++) {*c_out++=0;}
	  i+=delta;
	  end+=width*delta;
	  break;
	default: /* Literal */
	  if (rle==8){
	    for(j=0;(j<code)&&(c_out<=end);j++) {*c_out++=*c_in++;}
	    if (code&1) c_in++; /* Keep word aligned */
	  }
	  else{
	    for(j=0;(j<code/2)&&(c_out<=end);j++){
	      *c_out++=(*c_in&0xf0)>>4;
	      *c_out++=(*c_in++)&0xf;
	    }
	    if (code&1) *c_out++=((*c_in++)&0xf0)>>4;
	    if ((code+1)&2) c_in++; /* Keep word aligned */
	  }
	}
      }
    }
    if (eol){
      fprintf(stderr,"Bad line length in RLE decode line=%d\n",i);
      exit(1);
    }
    /* Should now have an EOL or EOD code */
    if (*(c_in++)!=0){
      fprintf(stderr,"Bad line length in RLE decode line=%d\n",i);
      exit(1);
    }
    if ((*c_in!=0)&&((*c_in!=1)&&(i=height))){
      fprintf(stderr,"Bad line length in RLE decode line=%d\n",i);
      exit(1);
    }
    c_in++;
  }

  image->data=out;
}
