/* Copyright (c) 2010 MJ Rutter 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street,
 * Fifth Floor, Boston, MA  02110-1301, USA.
 */ 

#include <stdio.h>  /* bmp2eps.h needs this... */

#include "bmp2eps.h"

/* We avoid using math.h / libm throughout, so don't use ceil() here */
#define iceil(x) ((int)(x+0.9999))

static void concat(double ctm[6], double trans[6]);

/* Work out transform matrix required for correct scaling and translation
   of image. Identity for EPS, [iw 0 0 ih 0 0] for PDF, and something
   complicated if a fit-to-page option has been given, or if landscape
   requested. */

void p_trans(struct bitmap *image, int pt,
	     int bb[4], double tm[6], int paper[2]){
  int ex,land,i,debug;
  unsigned int width,height;
  double tm2[6],scx,scy,sc,trx,try,iw,ih,scx2,scy2;
  double a0=3370;  /* Points in length of A0 */

  debug=image->flags&DEBUG_MASK;
  width=image->width;
  height=image->height;

  tm[0]=tm[3]=1;
  tm[1]=tm[2]=tm[4]=tm[5]=0;
  scx=scy=1;
  trx=try=0;

  land=(image->flags&LAND);
  ex=image->flags&EXPAND;

  if (debug>2) fprintf(stderr,"BoundingBox on entry to p_trans: %d %d %d %d\n",
                               bb[0],bb[1],bb[2],bb[3]);

  if (bb[2]-bb[0]>0){
    if (bb[3]-bb[1]==0){
      if (!land) bb[3]=(bb[2]-bb[0])*height/width+bb[1];
      else bb[3]=(bb[2]-bb[0])*width/height+bb[1];
    }
    if (!land){
      scx=(bb[2]-bb[0])/(double)width;
      scy=(bb[3]-bb[1])/(double)height;
      iw=bb[2]-bb[0];
      ih=bb[3]-bb[1];
    }else{
      scx=(bb[2]-bb[0])/(double)height;
      scy=(bb[3]-bb[1])/(double)width;
      ih=bb[2]-bb[0];
      iw=bb[3]-bb[1];
    }
    trx=bb[0];
    try=bb[1];
  }else{

    if ((image->dpi_x>0)&&(image->dpi_y>0)){
      scx=72./image->dpi_x;
      scy=72./image->dpi_y;
    }

    if (pt>=0){  /* ISO A0 series */
      for(i=0;i<pt;i++) a0=a0/1.414213562373;
      paper[1]=a0;
      paper[0]=a0/1.414213562373;
    }else{        /* US letter */
      paper[0]=612;
      paper[1]=792;
    }
    
    
    if (!land){
      iw=width*scx;
      ih=height*scy;
    }else{
      iw=height*scy;
      ih=width*scx;
    }
    
    if (!(image->flags&PAGE)){
      bb[0]=bb[1]=0;
      bb[2]=iceil(iw);
      bb[3]=iceil(ih);
    }else{
      scx2=((double)(0.9*paper[0]))/iw;
      scy2=((double)(0.9*paper[1]))/ih;
      sc=(scx2<scy2) ? scx2 : scy2;
      if((sc>1)&&!ex) sc=1;
      
      scx*=sc;
      scy*=sc;
      iw*=sc;
      ih*=sc;
      
      trx=(paper[0]-iw)/2;
      try=(paper[1]-ih)/2;
      bb[0]=trx;
      bb[1]=try;
      bb[2]=trx+iw;
      bb[3]=try+ih;
    }
  }

  if (land){
    tm2[0]=tm2[3]=tm2[4]=tm2[5]=0;    /* 90 rotate */
    tm2[1]=1;
    tm2[2]=-1;
    concat(tm,tm2);
    tm2[0]=tm2[3]=1;                  /* try (trx-bb[0]-bb[2]) translate */
    tm2[1]=tm2[2]=0;
    tm2[4]=try;
    tm2[5]=trx-bb[0]-bb[2];
    concat(tm,tm2);
  }else{
    tm2[0]=tm2[3]=1;                  /* trx try translate */
    tm2[1]=tm2[2]=0;
    tm2[4]=trx;
    tm2[5]=try;
    concat(tm,tm2);
  }

  tm2[0]=scx; tm2[3]=scy;             /* scx scy scale */
  tm2[1]=tm2[2]=tm2[4]=tm2[5]=0;  
  concat(tm,tm2);

  /* Whereas PostScript's image operator draws with unit square pixels,
   * PDF's draws into the unit square. Hence PDF needs an extra
   * width height scale
   */

  if (image->flags&PDF){
    tm2[0]=width;
    tm2[3]=height;
    tm2[1]=tm2[2]=tm2[4]=tm2[5]=0;
    concat(tm,tm2);
  }

  if (debug>2){
    fprintf(stderr,"BoundingBox on exit: %d %d %d %d\n",
            bb[0],bb[1],bb[2],bb[3]);
    fprintf(stderr,"tm on exit: %f %f %f %f %f %f\n",
            tm[0],tm[1],tm[2],tm[3],tm[4],tm[5]);
  }
}

/* Emulate the PS concat operator */
static void concat(double ctm[6], double trans[6]){
  int i;
  double new[6];

  new[0]=trans[0]*ctm[0]+trans[1]*ctm[2];
  new[1]=trans[0]*ctm[1]+trans[1]*ctm[3];

  new[2]=trans[2]*ctm[0]+trans[3]*ctm[2];
  new[3]=trans[2]*ctm[1]+trans[3]*ctm[3];

  new[4]=trans[4]*ctm[0]+trans[5]*ctm[2]+ctm[4]; 
  new[5]=trans[4]*ctm[1]+trans[5]*ctm[3]+ctm[5]; 

  for(i=0;i<6;i++) ctm[i]=new[i];
}
