/* Copyright (c) 2012 MJ Rutter 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street,
 * Fifth Floor, Boston, MA  02110-1301, USA.
 */ 


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "bmp2eps.h"

void print_bitmap(struct bitmap *image){
  int debug;
  int i,t;

  debug=image->flags&DEBUG_MASK;

  if (!debug) return;

  fprintf(stderr,"Image %dx%d depth %d with %d components\n",
	  image->width,image->height,image->depth,image->components);
  fprintf(stderr,"DPI: %dx%d\n",image->dpi_x,image->dpi_y);

  if (image->data)
    fprintf(stderr,"Image data present length %u\n",image->data_len);
  else
    fprintf(stderr,"Image data not present\n");

  fprintf(stderr,"Compression / compression requested ");
  switch(image->compression){
  case 0:
    fprintf(stderr,"none\n");
    break;
  case RLE:
    fprintf(stderr,"RLE\n");
    break;
  case LZW:
    fprintf(stderr,"LZW\n");
    break;
  case CCITT:
    fprintf(stderr,"CCITT\n");
    break;
  case FLATE:
    fprintf(stderr,"FLATE\n");
    break;
  case RLE24:
    fprintf(stderr,"RLE24\n");
    break;
  case DCT:
    fprintf(stderr,"DCT\n");
    break;
  case CCITTG4:
    fprintf(stderr,"CCITTG4\n");
    break;
  case FLATEF:
    fprintf(stderr,"FLATE (filtered strategy)\n");
    break;
  case PNG:
    fprintf(stderr,"PNG (passthrough)\n");
    break;
  case PNG_A7:
    fprintf(stderr,"interlaced PNG (passthrough)\n");
    break;
  default:
    fprintf(stderr,"unexpected type %d\n",image->compression);
  }

  if (image->palette)
    fprintf(stderr,"Image palette present length %d\n",image->pal_len);
  else
    fprintf(stderr,"Image palette not present\n");

  fprintf(stderr,"Image type is ");
  switch(image->type){
  case BMP:
    fprintf(stderr,"BMP");
    break;
  case PALETTE:
    fprintf(stderr,"PALETTE");
    break;
  case DCT:
    fprintf(stderr,"DCT");
    break;
  default:
    fprintf(stderr,"unexpected type %d",image->type);
  }
  fprintf(stderr,"\n");

  fprintf(stderr,"Image colour is ");
  switch(image->colour){
  case GREY:
    fprintf(stderr,"GREY");
    break;
  case RGB:
    fprintf(stderr,"RGB");
    break;
  case RGBA:
    fprintf(stderr,"RGBA");
    break;
  case GREYA:
    fprintf(stderr,"GREYA (grey plus alpha)");
    break;
  case A1RGB:
    fprintf(stderr,"A1RGB (0x00/0xff alpha plus RGB)");
    break;
  case A1GREY:
    fprintf(stderr,"A1GREY (0x00/0xff alpha plus grey)");
    break;
  case PALETTE:
    fprintf(stderr,"PALETTE");
    break;
  default:
    fprintf(stderr,"unexpected colour %d",image->colour);
  }
  fprintf(stderr,"\n");

  if (image->trans){
    fprintf(stderr,"Transparent colour %2x",image->trans[0]);
    if ((image->colour==RGB)||(image->colour==RGBA))
      fprintf(stderr,"%2x%2x",image->trans[1],image->trans[2]);
    fprintf(stderr,"\n");
  }
  else fprintf(stderr,"Transparent colour absent\n");

  if (image->apalette){
    fprintf(stderr,"Alpha palette of length %d present\n",image->apal_len);
    t=0;
    for(i=0;i<image->apal_len;i++)
      if (image->apalette[i]==255) t++;
    fprintf(stderr,"  %d entries opaque\n",t);
  }

  if ((image->palette)&&(debug>2)){
    fprintf(stderr,"Palette:\n");
    for(i=0;i<image->pal_len;i++){
      fprintf(stderr,"%3d %2x %2x %2x",i,image->palette[3*i],
	      image->palette[3*i+1],image->palette[3*i+2]);
      if (image->apalette){
	if (i<image->apal_len)
	  fprintf(stderr," %2x\n",image->apalette[i]);
	else
	  fprintf(stderr," ff\n");
      }
      else fprintf(stderr,"\n");
    }
  }
	  
  if (image->bkgd){
    fprintf(stderr,"Background colour %02x",image->bkgd[1]);
    if (image->bkgd[0]==3)
      fprintf(stderr,"%02x%02x",image->bkgd[2],image->bkgd[3]);
    fprintf(stderr,"\n");
  }
  else fprintf(stderr,"Background colour absent\n");

  fprintf(stderr,"Soft mask %s\n",(image->smask)?"present":"absent");

  fprintf(stderr,"Flags: %8x\n",image->flags);

}
