/* Copyright (c) 2010-17 MJ Rutter 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street,
 * Fifth Floor, Boston, MA  02110-1301, USA.
 */ 

/* Adobe's RLE and TIFF's packbits are identical, save for the
 * EOD marker in Adobe's version. HP RTL uses the TIFF version.
 */

/* 6/2017 Correct read of one byte after end of input buffer */

unsigned to_rle(unsigned char *raw, unsigned char *rle, unsigned len){
  unsigned char *a,*b,*c,*end;

  end=raw+len-1;

  a=raw;
  c=rle;
  while(a<=end){
    b=a+1;
    while ((b-a<128)&&(b<=end)&&(*b==*a)) b++;
    b--;
    if (b-a>1){ /* A run */
      *(c++)=(unsigned char)(256-(b-a));
      *(c++)=*a;
      a=b+1;
    }else{ /* Literal bytes */
      b++;
      while ((b-a<128)&&(b<=end)&&
             !((*b==*(b-1))&&((b<end)?(*b==*(b+1)):0))) b++;
      b--;
      if (b<end) b--;
      *(c++)=(unsigned char)(b-a);
      while(a<=b) *(c++)=*(a++);
    }
  }

  *(c++)=0x80; /* Adobe's EOD marker */

  return(c-rle);
}

#if 0
/* For debugging only. This uses HP's convention (0x80 is ignored, not EOD) */

unsigned int from_rle(char *in, char *out, int n){
  /* in - input buffer
   * out - output buffer
   * n - length of input buffer
   *
   * output buffer must be big enough!
   *
   * returns length of output buffer
   */
  int i,count;
  char *p1,*p2;

  p1=in;
  p2=out;
  
  while(p1<in+n){
    if ((*p1)>=0){ /* literal bytes */
      count=*(p1++);
      count++;
      for(i=0;i<count;i++) *(p2++)=*(p1++);
    }
    else if ((*p1)==-128) p1++;
    else{
      if ((*p1)==-128){p1++;continue;}
      count=*(p1++);
      count=1-count;
      for(i=0;i<count;i++) *(p2++)=*p1;
      p1++;
    }
  }
  return (unsigned int)(p2-out);
}

#endif
