/* Copyright (c) 2010-2020 MJ Rutter 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street,
 * Fifth Floor, Boston, MA  02110-1301, USA.
 */ 

#include<stdio.h>
#include<stdlib.h>  /* exit() */
#include<stdarg.h>
#include <time.h>

#include"bmp2eps.h"

static FILE *op; 
static int pdf_p(char *fmt,...){
  va_list args;
  int i;
  va_start(args,fmt);
  i=vfprintf(op,fmt,args);
  va_end(args);
  return(i);
}

#define OBJ_PRN cref++;xref=realloc(xref,(cref+1)*sizeof(int)); \
                xref[cref]=xref[cref-1]+fprintf(out,"%d 0 obj\n",cref); \
                xref[cref]+=pdf_p

unsigned int A85p(FILE *out, unsigned char *in, int len);
unsigned int hexp(FILE *out, unsigned char* in,int len);
void p_trans(struct bitmap *image, int pt,int bb[4], double tm[6],
             int paper[2]);
int compress_final(unsigned char *data, unsigned char *out,
		   unsigned int data_len, int out_len,
		   int strategy);
void lzw_flate_set_level(int level);


void write_pdf(FILE *out,struct bitmap *image, int bb_in[4], int pt){
  int i,len,smooth,binary,hex,bb[4];
  int pdf_image_type,root,zlen,compress;
  char image_type[]="/ImageX /ImageC";
  unsigned char *zpal;
  double tm[6];
  static int nobjs,page_count=0,*xref,cref=0,*resources,*contents,*mbox;
  static int flags,paper[2];
  char timestr[40];
  time_t tim;
  struct tm *tm_tim;
  
  zpal=NULL;
  
  if (page_count==0){
    if (!image) error_exit("First call to write_pdf with null image");
    xref=malloc(sizeof(int));
    if (!xref) error_exit("Malloc for an int failed!");
    xref[0]=fprintf(out,"%%PDF-1.4\n");
    resources=contents=mbox=NULL;
    flags=image->flags;
  }

  if (!image){ /* finish PDF file */

    root=cref+page_count+1;
    
    /* leaves*/
    for(i=0;i<page_count;i++){
      OBJ_PRN("<</Type /Page\n"
	      "/Parent %d 0 R\n"
	      "/Resources %d 0 R\n"
	      "/Contents %d 0 R\n",root,resources[i],contents[i]);
      len=0;
      if (!(flags&PAGE))
	len+=fprintf(out,"/MediaBox [0 0 %d %d]\n",mbox[2*i],mbox[2*i+1]);
      len+=fprintf(out,">>\nendobj\n\n");
      xref[cref]+=len;
    }

    /* root */
    OBJ_PRN("<</Type /Pages\n");
    len=0;
    /* Is MediaBox constant? */
    if (flags&PAGE)
      len+=fprintf(out,"/MediaBox [0 0 %d %d]\n",paper[0],paper[1]);
    len+=fprintf(out,"/Count %d\n/Kids [",page_count);
    len+=fprintf(out,"%d 0 R",cref-page_count);
    for(i=1;i<page_count;i++)
      len+=fprintf(out,"\n      %d 0 R",cref-page_count+i);
    len+=fprintf(out,"]\n"
            ">>\nendobj\n\n");
    xref[cref]+=len;
    
    /* catalogue */
    OBJ_PRN("<</Type /Catalog /Pages %d 0 R>>\nendobj\n\n",cref-1);

    /* info */
    OBJ_PRN("<</Producer (bmp2pdf " BMP2EPS_VER ")");
    len=0;

    tim=time(NULL);
    tm_tim=localtime(&tim);
    strftime(timestr,40,"%Y%m%d%H%M",tm_tim);
    len+=fprintf(out," /CreationDate (D:%s)",timestr);

    len+=fprintf(out,">>\nendobj\n\n");
    xref[cref]+=len;
    
    nobjs=cref+1;
    fprintf(out,"xref\n0 %d\n0000000000 65535 f \n",nobjs);

    for(i=0;i<nobjs-1;i++) fprintf(out,"%010d 00000 n \n",xref[i]);
    
    fprintf(out,"trailer\n<<\n/Size %d\n/Root %d 0 R\n/Info %d 0 R\n>>\n",
            nobjs+1,root+1,root+2);

    fprintf(out,"startxref\n%d\n%%%%EOF\n",xref[nobjs-1]);

    if (mbox) free(mbox);
    free(resources);
    free(contents);
    free(xref);
    return;
  }

  
  page_count++;
  for(i=0;i<4;i++) bb[i]=bb_in[i];
  op=out;

  if((image->compression==RLE)&&(image->depth!=8)){
    fprintf(stderr,"Unable to pdf_write RLE except at 8 bit\n");
    exit(1);
  }

  p_trans(image,pt,bb,tm,paper);

  if (!(image->flags&PAGE)){
    mbox=realloc(mbox,2*page_count*sizeof(int));
    if (!mbox) error_exit("realloc error for mbox");
    mbox[2*page_count-2]=bb[2];
    mbox[2*page_count-1]=bb[3];
  }
  
  smooth=(image->flags&SMOOTH)?1:0;
  binary=(image->flags&BINARY)?1:0;
  hex=(image->flags&HEX)?1:0;
  if (image->flags&HEX2) hex+=2;

  pdf_image_type='C';
  if (image->colour==GREY)  pdf_image_type='B';
  if (image->type==PALETTE) pdf_image_type='I';
  image_type[6]=pdf_image_type;
  if (pdf_image_type!='I') image_type[7]=0;

  
  /* resources dictionary */
  OBJ_PRN("<<\n"
          "  /XObject <</Img %d 0 R>>\n",cref+3);


  if (image->smask){
    i=cref+4;
    if (!binary) i++;
    if (pdf_image_type=='I'){
      i++;
      if (!binary) i++;
    }
    /*    xref[cref]+=fprintf(out," /Msk %d 0 R",i); */
  }

  xref[cref]+=fprintf(out,"  /ProcSet [/PDF %s]\n"
		      ">>\nendobj\n\n",
		      image_type);
  resources=realloc(resources,page_count*sizeof(int));
  resources[page_count-1]=cref;
  
  /* page contents */
  OBJ_PRN("<</Length %d 0 R>>\n"
          "stream\n",cref+1);

  len=fprintf(out,"q %g %g %g %g %g %g cm\n"
	      "/Img Do Q\n",tm[0],tm[1],tm[2],tm[3],tm[4],tm[5]);

  xref[cref]+=len+fprintf(out,"endstream\nendobj\n\n");
  contents=realloc(contents,page_count*sizeof(int));
  contents[page_count-1]=cref;

  OBJ_PRN("%d\nendobj\n\n",len);

  OBJ_PRN("<</Type /XObject\n"
          "/Subtype /Image\n/Width %d /Height %d\n"
          "/BitsPerComponent %d\n"
          "/ColorSpace ",
          image->width,image->height,
          (image->colour==RGB)?image->depth/3:image->depth);
  if (pdf_image_type=='B')
    xref[cref]+=fprintf(out,"/DeviceGray\n");
  else if (pdf_image_type=='C')
    xref[cref]+=fprintf(out,"/DeviceRGB\n");
  else
    xref[cref]+=fprintf(out,"[/Indexed /DeviceRGB %d %d 0 R]\n",
        	       image->pal_len-1,cref+((binary)?1:2));
  if (image->trans){
    xref[cref]+=fprintf(out,"/Mask [");
    if (image->colour==RGB) xref[cref]+=fprintf(out,"%d %d %d %d %d %d]\n",
                                                      (int)image->trans[0],
                                                      (int)image->trans[0],
                                                      (int)image->trans[1],
                                                      (int)image->trans[1],
                                                      (int)image->trans[2],
                                                      (int)image->trans[2]);
    else xref[cref]+=fprintf(out,"%d %d]\n",(int)(*image->trans),
                                            (int)(*image->trans));
  }
  if (image->smask) xref[cref]+=fprintf(out,"/SMask %d 0 R\n",i);
  if (smooth) xref[cref]+=fprintf(out,"/Interpolate true\n");
  xref[cref]=xref[cref]+fprintf(out,"/Filter [");
  if (!binary) xref[cref]=xref[cref]+fprintf(out,"/ASCII85Decode ");

  if (image->compression==DCT)
    xref[cref]+=fprintf(out,"/DCTDecode]");
#ifdef USE_LZW
  else if (image->compression==LZW){
    xref[cref]+=fprintf(out,"/LZWDecode]\n");
    if (image->flags&PREDICT) {
      xref[cref]+=fprintf(out,"/DecodeParms [%s\n"
			  "<</Predictor 10 "
			  "/Columns %d /Colors %d "
			  "/BitsPerComponent %d>>\n]",
			  (binary)?"":"null ",
			  image->width,
			  (image->colour==RGB)?3:1,
			  (image->colour==RGB)?image->depth/3:
			  image->depth);
    }
  }
#endif
  else if (image->compression==FLATE){
    xref[cref]+=fprintf(out,"/FlateDecode]");
    if (image->flags&PREDICT) {
      xref[cref]+=fprintf(out,"/DecodeParms [%s\n"
			  "<</Predictor 10 "
			  "/Columns %d /Colors %d "
			  "/BitsPerComponent %d>>\n]",
			  (binary)?"":"null ",
			  image->width,
			  (image->colour==RGB)?3:1,
			  (image->colour==RGB)?image->depth/3:
			  image->depth);
    }
  }
  else if ((image->compression==CCITT)||(image->compression==CCITTG4)){
    xref[cref]+=fprintf(out,"/CCITTFaxDecode]\n"
        	       "/DecodeParms [%s<</Columns %d /Rows %d",
        	       (binary)?"":"null ",image->width,image->height);
    if (image->compression==CCITTG4) xref[cref]+=fprintf(out," /K -1");
    xref[cref]+=fprintf(out,">>]");
  }
  else if (image->compression==RLE)
    xref[cref]+=fprintf(out," /RunLengthDecode]");
  else if (image->compression==0)
    xref[cref]+=fprintf(out,"]");
  else{
    fprintf(stderr,"Unexpected compression type %d in write_pdf\n",
            image->compression);
    exit(1);
  }

  if (binary){
    xref[cref]=xref[cref]+fprintf(out,"\n"
        			 "/Length %d\n"
        			 ">> stream\n",image->data_len);
    len=fwrite(image->data,1,image->data_len,out);
  }
  else{
    xref[cref]=xref[cref]+fprintf(out,"\n"
        			 "/Length %d 0 R\n"
        			 ">> stream\n",cref+1);
    len=A85p(out,image->data,image->data_len);
  }

  xref[cref]=xref[cref]+len+fprintf(out,"\nendstream\nendobj\n\n");

  if (!binary) {OBJ_PRN("%d\nendobj\n\n",len);}

  if (pdf_image_type=='I'){
#ifdef USE_ZLIB
    if (binary){
      /* Should we compress the palette? */
      if (3*image->pal_len>50){
	zlen=3.03*image->pal_len+15;
	zpal=malloc(zlen);
	if (!zpal) error_exit("malloc error in write_pdf");
#ifdef USE_ZOPFLI
	compress=ZOPFLI;
#elif defined USE_ZLIB
	compress=FLATE;
#endif
	lzw_flate_set_level(image->compress_level);
	zlen=compress_final(image->palette,zpal,3*image->pal_len,zlen,compress);
	if (zlen>3*image->pal_len-40){
	  free(zpal);
	  zpal=NULL;
	}
      }
#endif
      if (zpal){
	OBJ_PRN("<</Filter [/FlateDecode] /Length %d>>\nstream\n",zlen);
	xref[cref]+=fwrite(zpal,1,zlen,out);
	free(zpal);
      }
      else{
	OBJ_PRN("<</Length %d\n>>\nstream\n",3*image->pal_len);
	xref[cref]+=fwrite(image->palette,1,3*image->pal_len,out);
      }
      xref[cref]+=fprintf(out,"\nendstream\nendobj\n\n");
    }
    else{
      OBJ_PRN("<</Filter %s /Length %d 0 R>>\nstream\n",
              (hex==0)?"/ASCII85Decode":"/ASCIIHexDecode",cref+1);
      if (hex==0)
        len=A85p(out,image->palette,3*image->pal_len);
      else
        len=hexp(out,image->palette,3*image->pal_len);
      xref[cref]+=len+fprintf(out,"endstream\nendobj\n\n");
      OBJ_PRN("%d\nendobj\n\n",len);
    }
  }

  if (image->smask){
    OBJ_PRN("<</Type /XObject\n"
          "/Subtype /Image\n/Width %d /Height %d\n"
          "/BitsPerComponent 8\n"
          "/ColorSpace /DeviceGray\n",
          image->smask->width,image->smask->height);
    if (smooth) xref[cref]+=fprintf(out,"/Interpolate true\n");
    xref[cref]=xref[cref]+fprintf(out,"/Filter [");
    if (!binary) xref[cref]=xref[cref]+fprintf(out,"/ASCII85Decode");
    if (image->smask->compression==FLATE){
      xref[cref]+=fprintf(out," /FlateDecode]\n");
      if (image->smask->flags&PREDICT) {
	xref[cref]+=fprintf(out,"/DecodeParms [%s\n"
			    "<</Predictor 10 "
			    "/Columns %d /Colors 1 "
			    "/BitsPerComponent 8>>\n]",
			    (binary)?"":"null ",
			    image->smask->width);
      }
    }
#ifdef USE_LZW
    else if (image->smask->compression==LZW){
      xref[cref]+=fprintf(out," /LZWDecode]\n");
      if (image->smask->flags&PREDICT) {
	xref[cref]+=fprintf(out,"/DecodeParms [%s\n"
			    "<</Predictor 10 "
			    "/Columns %d /Colors 1 "
			    "/BitsPerComponent 8>>\n]",
			    (binary)?"":"null ",
			    image->smask->width);
      }
    }
#endif
    else if (image->smask->compression==RLE)
      xref[cref]+=fprintf(out," /RunLengthDecode]");
    else if (image->smask->compression==0)
      xref[cref]+=fprintf(out,"]");
    else{
      fprintf(stderr,"Unexpected smask compression type %d in write_pdf\n",
	image->smask->compression);
      exit(1);
    }
    
    if (binary){
      xref[cref]=xref[cref]+fprintf(out,"\n"
	"/Length %d\n"
	">> stream\n",image->smask->data_len);
      len=fwrite(image->smask->data,1,image->smask->data_len,out);
    }
    else{
      xref[cref]=xref[cref]+fprintf(out,"\n"
	"/Length %d 0 R\n"
	">> stream\n",cref+1);
      len=A85p(out,image->smask->data,image->smask->data_len);
    }

    xref[cref]=xref[cref]+len+fprintf(out,"endstream\nendobj\n\n");

    if (!binary) {OBJ_PRN("%d\nendobj\n\n",len);}

  } /* end if (image->smask) */


}
