/* Copyright (c) 2010-2019 MJ Rutter 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street,
 * Fifth Floor, Boston, MA  02110-1301, USA.
 */ 

#include<stdio.h>
#include<stdlib.h>  /* exit() */
#include<stdarg.h>

#include"bmp2eps.h"

unsigned int crc_table[256];

void init_crc_table(void);
unsigned int crc(unsigned char *buf, int len);
unsigned int update_crc(unsigned int crc, unsigned char *buf, int len);

void write_png(FILE *out,struct bitmap *image){
  unsigned char buff[1024];
  unsigned int ccrc,len;
  int debug,i;

  if((image->compression!=FLATE)&&(image->compression!=PNG_A7)){
    fprintf(stderr,"Unable to write non flate compressed images as PNG.\n");
    exit(1);
  }

  debug=image->flags&DEBUG_MASK;

  if (debug)
    fprintf(stderr,"Writing PNG image %dx%d depth %d\n",
	    image->width,image->height,image->depth);
  if (debug>1) fprintf(stderr,"%d components\n",image->components);

  init_crc_table();

  /* Write PNG header */

  buff[0]=0x89;
  buff[1]='P';
  buff[2]='N';
  buff[3]='G';
  buff[4]=13;
  buff[5]=10;
  buff[6]=26;
  buff[7]=10;
  fwrite(buff,1,8,out);

  /* IHDR chunk */

  buff[0]=buff[1]=buff[2]=0;
  buff[3]=13;
  buff[4]='I';
  buff[5]='H';
  buff[6]='D';
  buff[7]='R';

  buff[8]=(image->width&0xff000000u)>>24;
  buff[9]=(image->width&0xff0000)>>16;
  buff[10]=(image->width&0xff00)>>8;
  buff[11]=image->width&0xff;

  buff[12]=(image->height&0xff000000u)>>24;
  buff[13]=(image->height&0xff0000)>>16;
  buff[14]=(image->height&0xff00)>>8;
  buff[15]=image->height&0xff;

  buff[16]=image->depth/image->components;

  buff[17]=((image->colour==PALETTE)?1:0)+
    ((image->colour==GREY)?0:2);
  if (image->colour==GREYA) buff[17]=4;
  if (image->colour==RGBA) buff[17]=6;

  buff[18]=0;  /* compression type */

  buff[19]=0;  /* filter type */

  buff[20]=0;  /* interlace type (none) */
  if (image->compression==PNG_A7) buff[20]=1;

  ccrc=crc(buff+4,17);

  buff[21]=(ccrc&0xff000000u)>>24;
  buff[22]=(ccrc&0xff0000)>>16;
  buff[23]=(ccrc&0xff00)>>8;
  buff[24]=ccrc&0xff;
  fwrite(buff,1,25,out);

  /* PLTE chunk */ 

  if (image->colour==PALETTE){
    len=image->pal_len*3;
    buff[0]=(len&0xff000000u)>>24;
    buff[1]=(len&0xff0000)>>16;
    buff[2]=(len&0xff00)>>8;
    buff[3]=len&0xff;
    buff[4]='P';
    buff[5]='L';
    buff[6]='T';
    buff[7]='E';
    fwrite(buff,1,8,out);
    ccrc=0xffffffffu;
    ccrc=update_crc(ccrc,buff+4,4);
    fwrite(image->palette,1,len,out);
    ccrc=update_crc(ccrc,image->palette,len);
    ccrc^=0xffffffffu;
    buff[0]=(ccrc&0xff000000u)>>24;
    buff[1]=(ccrc&0xff0000)>>16;
    buff[2]=(ccrc&0xff00)>>8;
    buff[3]=ccrc&0xff;
    fwrite(buff,1,4,out);
  }

  /* tRNS chunk */

  if ((image->trans)||(image->apalette)){
    len=2;
    if (image->colour==RGB) len=6;
    if (image->colour==PALETTE) {
      if (image->apalette) len=image->apal_len;
      else len=(unsigned int)(*image->trans)+1;
    }
    buff[0]=(len&0xff000000u)>>24;
    buff[1]=(len&0xff0000)>>16;
    buff[2]=(len&0xff00)>>8;
    buff[3]=len&0xff;
    buff[4]='t';
    buff[5]='R';
    buff[6]='N';
    buff[7]='S';
    fwrite(buff,1,8,out);
    ccrc=0xffffffffu;
    ccrc=update_crc(ccrc,buff+4,4);
    if (image->colour==RGB){
      buff[0]=0;
      buff[1]=image->trans[0];
      buff[2]=0;
      buff[3]=image->trans[1];
      buff[4]=0;
      buff[5]=image->trans[2];
    }
    else if ((image->colour==PALETTE)&&(!image->apalette)){
      for(i=0;i<image->trans[0];i++) buff[i]=255;
      buff[image->trans[0]]=0;
    }
    else if (image->colour==GREY){
      buff[0]=0;
      buff[1]=image->trans[0];
    }
    else if(!((image->colour==PALETTE)&&(image->apalette))){
      fprintf(stderr,"Confused re transparency in write_png\n");
      exit(1);
    }
    if ((image->colour==PALETTE)&&(image->apalette)){
      fwrite(image->apalette,1,len,out);
      ccrc=update_crc(ccrc,image->apalette,len);
    }
    else{
      fwrite(buff,1,len,out);
      ccrc=update_crc(ccrc,buff,len);
    }
    ccrc^=0xffffffffu;
    buff[0]=(ccrc&0xff000000u)>>24;
    buff[1]=(ccrc&0xff0000)>>16;
    buff[2]=(ccrc&0xff00)>>8;
    buff[3]=ccrc&0xff;
    fwrite(buff,1,4,out);
  }

  /* pHYs chunk */

  if ((image->dpi_x>0)&&(image->dpi_y>0)){
    buff[0]=buff[1]=buff[2]=0;
    buff[3]=9;
    buff[4]='p';
    buff[5]='H';
    buff[6]='Y';
    buff[7]='s';
    len=image->dpi_x*1000/25.4;
    buff[8]=(len&0xff000000u)>>24;
    buff[9]=(len&0xff0000)>>16;
    buff[10]=(len&0xff00)>>8;
    buff[11]=len&0xff;
    len=image->dpi_y*1000/25.4;
    buff[12]=(len&0xff000000u)>>24;
    buff[13]=(len&0xff0000)>>16;
    buff[14]=(len&0xff00)>>8;
    buff[15]=len&0xff;
    buff[16]=1;
    ccrc=crc(buff+4,13);
    buff[17]=(ccrc&0xff000000u)>>24;
    buff[18]=(ccrc&0xff0000)>>16;
    buff[19]=(ccrc&0xff00)>>8;
    buff[20]=ccrc&0xff;
    fwrite(buff,1,21,out);
  }

  /* bKGD chunk  - write only if transparency info exists */
  if (((image->trans)||(image->colour==RGBA)||(image->colour==GREYA))&&
        (image->bkgd)){
    buff[0]=buff[1]=buff[2]=buff[3]=0;
    buff[4]='b';
    buff[5]='K';
    buff[6]='G';
    buff[7]='D';
    if (image->colour==PALETTE){
      buff[3]=1;
      buff[8]=image->bkgd[1];
    }
    else if ((image->colour==GREY)||(image->colour==GREYA)){
      buff[3]=2;
      buff[8]=0;
      buff[9]=image->bkgd[1];
    }
    else if ((image->colour==RGB)||(image->colour==RGBA)){
      buff[3]=6;
      buff[8]=0;
      buff[9]=image->bkgd[1];
      buff[10]=0;
      buff[11]=image->bkgd[2];
      buff[12]=0;
      buff[13]=image->bkgd[3];
    }
    if (buff[3]){
      ccrc=crc(buff+4,buff[3]+4);
      i=buff[3]+8;
      buff[i]=(ccrc&0xff000000u)>>24;
      buff[i+1]=(ccrc&0xff0000)>>16;
      buff[i+2]=(ccrc&0xff00)>>8;
      buff[i+3]=ccrc&0xff;
      fwrite(buff,1,i+4,out);
    }
  }

  /* IDAT chunk */

  buff[0]=(image->data_len&0xff000000u)>>24;
  buff[1]=(image->data_len&0xff0000)>>16;
  buff[2]=(image->data_len&0xff00)>>8;
  buff[3]=image->data_len&0xff;
  buff[4]='I';
  buff[5]='D';
  buff[6]='A';
  buff[7]='T';
  fwrite(buff,1,8,out);
  ccrc=0xffffffffu;
  ccrc=update_crc(ccrc,buff+4,4);
  fwrite(image->data,1,image->data_len,out);
  ccrc=update_crc(ccrc,image->data,image->data_len);
  ccrc^=0xffffffffu;
  buff[0]=(ccrc&0xff000000u)>>24;
  buff[1]=(ccrc&0xff0000)>>16;
  buff[2]=(ccrc&0xff00)>>8;
  buff[3]=ccrc&0xff;
  fwrite(buff,1,4,out);

  /* IEND chunk */

  buff[0]=buff[1]=buff[2]=buff[3]=0;
  buff[4]='I';
  buff[5]='E';
  buff[6]='N';
  buff[7]='D';
  ccrc=crc(buff+4,4);
  buff[8]=(ccrc&0xff000000u)>>24;
  buff[9]=(ccrc&0xff0000)>>16;
  buff[10]=(ccrc&0xff00)>>8;
  buff[11]=ccrc&0xff;
  fwrite(buff,1,12,out);  

}

/* CRC code following sample code in PNG specification */

/* make table for fast crc lookups */
void init_crc_table(void)
{
  unsigned int c,n;
  int k;
 
  for(n=0;n<256;n++){
    c=n;
    for(k=0;k<8;k++)
      c=(c&1)?(0xedb88320u^(c>>1)):(c>>1);
    crc_table[n]=c;
  }
}

/* update a running crc */

unsigned int update_crc(unsigned int crc, unsigned char *buf, int len)
{
  unsigned int c=crc;
  unsigned char *p=buf;
  int n=len;

  if (n>0)
    do{
      c=crc_table[(c^(*p++))&0xff]^(c>>8);
    } while (--n);
  return c;
}

/* return the crc of buffer in form suitable for storage,
   i.e. xored with ~0 */
unsigned int crc(unsigned char *buf, int len)
{
  return update_crc(~0, buf, len)^~0;
}
