/* Copyright (c) 2010-2020 MJ Rutter 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street,
 * Fifth Floor, Boston, MA  02110-1301, USA.
 */ 

/* May 2020 -- add multi-page PS output */

#include <stdio.h>
#include <stdlib.h>

#include "bmp2eps.h"

unsigned int A85p(FILE *out, unsigned char *in, int len);
unsigned int hexp(FILE *out, unsigned char *in, int len);
unsigned to_rle(unsigned char *raw,unsigned char *rle, unsigned len);
void p_trans(struct bitmap *image, int pt,int bb[4], double tm[6],
             int paper[2]);
int compress_final(unsigned char *data, unsigned char *out,
		   unsigned int data_len, int out_len,
		   int strategy);
void lzw_flate_set_level(int level);

void write_ps(FILE *out, struct bitmap *image, int bb_in[4], int pt)
{
  int i,j,k,distill,ps,bb[4];
  int compress,dict=0,hex,land,smooth,binary;
  int ncols,paper[2];
  unsigned int width,height,scanlen;
  double tm[6];

  unsigned char *scan1,*scan2,*p;
  unsigned char *col;

  static int page_count=0;

  if (!image){ /* finish PS file */
    fprintf(out,"%%%%Trailer\n%%%%Pages: %d\n%%%%EOF\n",page_count);
    return;
  }
  
  compress=image->compression;
  width=image->width;
  height=image->height;
  ps=(image->flags&PAGE)?1:0;
  distill=(image->flags&DISTILL)?1:0;

  for(i=0;i<4;i++) bb[i]=bb_in[i];
  p_trans(image,pt,bb,tm,paper);

  hex=0;
  if (image->flags&HEX) hex=1;
  if (image->flags&HEX2) hex+=2;
  land=(image->flags&LAND);
  smooth=image->flags&SMOOTH;
  binary=image->flags&BINARY;

  /* Now write the EPS */

  if (page_count==0){

    if (ps){
      fprintf(out,"%%!PS-Adobe-2.0\n"
	      "%%%%BoundingBox: 0 0 %d %d\n"
	      "%%%%LanguageLevel: 3\n",paper[0],paper[1]);
    }
    else{
      fprintf(out,"%%!PS-Adobe-2.0 EPSF-2.0\n"
	      "%%%%BoundingBox: %d %d %d %d\n"
	      "%%%%LanguageLevel: 3\n",
	      bb[0],bb[1],bb[2],bb[3]);
    }
    if (land) fprintf(out,"%%%%Orientation: Landscape\n");
    if (ps) fprintf(out,"%%%%Pages: (atend)\n");
    fprintf(out,"%%%%Creator: bmp2eps " BMP2EPS_VER "\n"
	    "%%%%DocumentData: %s\n"
	    "%%%%EndComments\n\n",(binary)?"8Bit":"Clean7Bit");

    if (distill){ /* save and set Distiller parameters */
      fprintf(out,"3 dict begin\n"
	      "/currentdistillerparams where\n"
	      " {pop} {/currentdistillerparams {1 dict} def} ifelse\n"
	      "/setdistillerparams where\n"
	      " {pop} {/setdistillerparams {pop} def} ifelse\n"
	      "/olddistillerparams currentdistillerparams def\n");

      if (image->depth==1){
	fprintf(out,"<< /DownSampleMonoImages false /EncodeMonoImages true\n"
		"   /AutoFilterMonoImages false /MonoImageFilter \n");
      }else if ((image->colour==GREY)||(image->colour==A1GREY)){
	fprintf(out,"<< /DownSampleGrayImages false /EncodeGrayImages true\n"
		"   /AutoFilterGrayImages false /GrayImageFilter \n");
      }else{
	fprintf(out,"<< /DownSampleColorImages false /EncodeColorImages true\n"
		"   /AutoFilterColorImages false /ColorImageFilter \n");
      }
      if (compress==DCT) fprintf(out,
				 "/DCTEncode /PassThroughJPEGImages true\n");
      else fprintf(out,"/FlateEncode\n");
    
      fprintf(out,">> setdistillerparams\n");
    }
  }
  
  if (!ps) fprintf(out,"gsave\n");
  else if (page_count==0)
    fprintf(out,"%%%%BeginSetup\n"
	    "<</PageSize [%d %d]>> setpagedevice\n"
	    "%%%%EndSetup\n\n",paper[0],paper[1]);

  page_count++;
  if (ps) {
    fprintf(out,"%%%%Page: %d %d\n",page_count,page_count);
    fprintf(out,"%%%%PageBoundingBox: %d %d %d %d\n",bb[0],bb[1],bb[2],bb[3]);
  }

  if ((image->colour==RGB)||
      (image->colour==A1RGB)) fprintf(out,"/DeviceRGB setcolorspace\n");
  else if ((image->colour==GREY)||
	   (image->colour==A1GREY)) fprintf(out,"/DeviceGray setcolorspace\n");
  else if (image->colour==PALETTE){
    ncols=image->pal_len;
    fprintf(out,"[/Indexed /DeviceRGB %d <",ncols-1);
    if (hex){
      col=image->palette;
      for(i=0;i<ncols;i+=8){
        fprintf(out,"\n");
        for(j=0;(j<8)&&(i+j<ncols);j++){
          fprintf(out,"%.2x%.2x%.2x ",(int)*col,(int)*(col+1),(int)*(col+2));
          col+=3;
        }
      }
      fprintf(out,">");
    }else{
      fprintf(out,"~\n");
      A85p(out,image->palette,3*ncols);
    }
    fprintf(out,"] setcolorspace\n\n");
  }

  if ((tm[1]==0)&&(tm[2]==0)&&(tm[4]==0)&&(tm[5]==0)){
    if ((tm[0]!=1)||(tm[3]!=1)) fprintf(out,"%g %g scale\n",tm[0],tm[3]);
  }
  else
    fprintf(out,"[%g %g %g %g %g %g] concat\n",tm[0],tm[1],tm[2],
            tm[3],tm[4],tm[5]);


  if ((image->flags&PSL2)&&((compress==FLATE)||(image->trans))){
    dict=1;
    fprintf(out,"1 dict begin\n"
           "languagelevel 3 lt {\n"
           " /image\n"
           "  {pop /Helvetica findfont 10 scalefont setfont"
           "   0 %d moveto 0 setgray\n"
           "   (This image requires level 3 PostScript) dup stringwidth pop\n"
           "   %d exch div 10 mul dup -.5 mul 0 exch rmoveto\n"
           "   /Helvetica findfont exch scalefont setfont show\n"
           "   0 0 %d %d rectstroke\n"
           "   currentfile ",
           image->height/2,image->width-4,
           image->width,image->height);
    /* currentfile /ASCII85Decode filter flushfile crashes our HP LJ4050
     * so... */
    if (binary)
      fprintf(out,"%d ()",image->data_len);
    else if (hex==2)
      fprintf(out,"0 (>)");
    else
      fprintf(out,"0 (~>)");
    fprintf(out," /SubFileDecode filter flushfile\n  }\n def} if\n");
  }

  if((image->colour==A1RGB)||(image->colour==A1GREY)){
    fprintf(out,"<</ImageType 3\n"
	    "  /InterleaveType 1\n"
            "  /MaskDict\n"
	    "  <</ImageType 1 /Width %d /Height %d\n"
	    "    /ImageMatrix [1 0 0 -1 0 %d]\n"
	    "    /BitsPerComponent %d /Decode [1 0]\n"
            "  >>\n"
	    "  /DataDict ",width,height,height,
	    image->depth/image->components);
  }

  if (compress!=RLE24){
/* ImageType must be 4 (a level 3 feature) if masking used */
    fprintf(out,"<</ImageType %d /Width %d /Height %d\n"
           "  /ImageMatrix [1 0 0 -1 0 %d]\n"
	    "  /DataSource currentfile",(image->trans)?4:1,
	    width,height,height);
    if (hex==2) fprintf(out," /ASCIIHexDecode filter\n");
    else if (!binary) fprintf(out," /ASCII85Decode filter\n");

    if (((compress==LZW)||(compress==FLATE))&&(image->flags&PREDICT)){
      fprintf(out,"  <</Predictor 10 "
	          "/Columns %d /Colors %d "
		  "/BitsPerComponent %d>>",
	          image->width,image->components,
  	          image->depth/image->components);
    }

    if (compress==LZW) fprintf (out,"  /LZWDecode filter\n");
    if (compress==FLATE) fprintf (out,"  /FlateDecode filter\n");
    if (compress==DCT) fprintf (out,"  /DCTDecode filter\n");
    if ((compress==CCITT)||(compress==CCITTG4)){
      fprintf (out,"  <</Columns %d /Rows %d",
              image->width,image->height);
      if (compress==CCITTG4) fprintf(out," /K -1");
      fprintf(out,">> /CCITTFaxDecode filter\n");
    }
    if (compress==RLE) fprintf (out,"  /RunLengthDecode filter\n");

    fprintf(out,"  /BitsPerComponent %d /Decode [0 %d",
           image->depth/image->components,
           image->type==PALETTE?((1<<image->depth)-1):1);
    if ((image->colour==RGB)||(image->colour==A1RGB)) fprintf(out," 0 1 0 1");
    fprintf(out,"]\n");

    if (image->trans){
      fprintf(out,"  /MaskColor [");
      if (image->colour==RGB) fprintf(out,"%d %d %d]\n",(int)image->trans[0],
                                                        (int)image->trans[1],
                                                        (int)image->trans[2]);
      else fprintf(out,"%d]\n",(int)(*image->trans));
    }
    if (smooth) fprintf(out,"  /Interpolate true\n");
    if ((image->colour==A1RGB)||(image->colour==A1GREY))  fprintf(out,">>\n");
    fprintf (out,">> image\n");
    if (hex==2) hexp(out,image->data,image->data_len);
    else if (!binary) A85p(out,image->data,image->data_len);
    else fwrite(image->data,1,image->data_len,out);
  }
  else if ((image->type==BMP)&&(image->colour==RGB)){
    scanlen=width;
    i=height;
    while((scanlen<=1024)&&(!(i&1))){scanlen<<=1;i>>=1;}
    dict=1;
    fprintf(out,"7 dict begin\n"
           "/rd %d string def /gn %d string def /bl %d string def\n"
           "/pixstr {currentfile /ASCII85Decode filter"
           " /RunLengthDecode filter} def\n"
           "/readr {pixstr rd readstring pop} def\n"
           "/readg {pixstr gn readstring pop} def\n"
           "/readb {pixstr bl readstring pop} def\n",
           scanlen,scanlen,scanlen);
  
    fprintf(out,"<</ImageType 1 /Width %d /Height %d\n"
           "  /ImageMatrix [1 0 0 -1 0 %d]\n"
           "  /BitsPerComponent 8 /Decode [0 1 0 1 0 1]\n",
           width,height,height);
    if (smooth) fprintf(out,"  /Interpolate true\n");
    fprintf(out,"  /MultipleDataSources true\n"
           "  /DataSource [ //readr //readg //readb ]"
           "\n>> image\n");

    scan1=malloc(scanlen);
    scan2=malloc(scanlen*51/50+1);
    if((!scan1)||(!scan2)){
      fprintf(stderr,"Malloc error in rle\n");
      exit(1);
    }
    p=image->data;
    for(p=image->data;(p-image->data)<image->data_len;p+=3*scanlen){
      /* red, green and blue */
      for(j=0;j<=2;j++){
        for(i=0;i<scanlen;i++) scan1[i]=*(p+3*i+j);
        k=to_rle(scan1,scan2,scanlen);
        A85p(out,scan2,k);
      }
    }
    free(scan2);
    free(scan1);
  }
  else{
    fprintf(stderr,"Impossible has happened in write_ps\n");
    exit(1);
  }

  fprintf(out,"\n");
  if (ps) fprintf(out,"showpage\n");
  if (dict) fprintf(out,"end\n");
  if (!ps) {
    fprintf(out,"grestore\n");
/* Restore Distiller parameters */
    if (distill) fprintf(out,"olddistillerparams setdistillerparams end\n");

    fprintf(out,"%%%%EOF\n");
  }

}
