/* Copyright (c) 2013-2017 MJ Rutter 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street,
 * Fifth Floor, Boston, MA  02110-1301, USA.
 */ 

/* 5/2017 - add compression for repeated rows in RGB output */

/* 6/2017 - add RLE compression, then seedrow */

/* 7/2018 - successfully tested on Z5600 (originally written for 1050c) */

#include<stdio.h>
#include<stdlib.h>  /* exit() */
#include<stdarg.h>
#include<string.h> /* memcmp() */

#include"bmp2eps.h"

static unsigned int seedrow(unsigned char *row, unsigned char *last,
                            unsigned char *out, unsigned int row_len);
#if DEBUG
static unsigned int unseedrow(unsigned char *in, unsigned char *last,
                              unsigned char *out, unsigned int data_len,
                              unsigned int row_len);
#endif
void write_rtl(FILE *out,struct bitmap *image){
  int debug,i,j,grey_step,row_len,dpi_x,dpi_y;
  unsigned int dup,pack,seed,c,s,written;
  unsigned char *p;
  unsigned char *rc,*rs;
#if DEBUG
  unsigned char *check;
#endif
  
  debug=image->flags&DEBUG_MASK;

  
  if ((image->depth!=1)&&(image->depth!=24)){
    if ((image->colour!=PALETTE)&&(image->colour!=GREY)){
      fprintf(stderr,"Unsupported depth in rtl\n");
      fprintf(stderr,"Depth %d colour %d\n",image->depth,image->colour);
      exit(1);
    }
  }

  if ((image->compression!=0)&&(image->compression!=RLE24)){
    fprintf(stderr,"Unsupported compression in rtl\n");
    exit(1);
  }

  if (image->width>36*600)
    fprintf(stderr,"*** WARNING: Image width > 21,600\n"
            "*** Image unlikely to print correctly on DJ1050\n");
  
  dpi_x=image->dpi_x;
  dpi_y=image->dpi_y;
  if (dpi_y==0) dpi_y=dpi_x;

  if (dpi_x==0){
    fprintf(stderr,"Warning: no DPI given or found\n");
    dpi_x=image->width/33;
    if (dpi_x<150) dpi_x=150;
    dpi_y=dpi_x;
    fprintf(stderr,"Setting DPI to %d\n",dpi_x);
  }
  
  dup=pack=seed=0;
  
  fprintf(out,"\x1b%%-12345X");
  fprintf(out,"@PJL JOB NAME=\"RTL\"\r\n");
  fprintf(out,"@PJL SET RESOLUTION=600\r\n");
  fprintf(out,"@PJL ENTER LANGUAGE=HPGL2\r\n");
     
  fprintf(out,"\033E");    /* reset */
  fprintf(out,"\x1b%%0B");  /* enter HPGL */
  fprintf(out,"BP5,1;");  /* Begin plot */
  fprintf(out,"PS%d,%d;",1+(image->height*1016)/dpi_y,
	                 1+(image->width*1016)/dpi_x);
  fprintf(out,"TR0;PU;");
  fprintf(out,"\x1b%%1A"); /* enter RTL */
  fprintf(out,"\x1b*v1N"); /* no source transparency */
  fprintf(out,"\x1b*v1O"); /* no pattern transparency */
  fprintf(out,"\x1b*p0x0Y"); /* cursor to origin */

  fprintf(out,"\x1b&a1N"); /* plot on the fly */
  fprintf(out,"\x1b*r%dS",image->width); /* source raster width (pixels) */
  fprintf(out,"\x1b*r%dT",image->height); /* source raster height (pixels) */
  if ((dpi_x==dpi_y)&&(600%dpi_x==0)){
    fprintf(out,"\x1b*t%dR",dpi_x); /* set DPI  */
    if (debug>1) fprintf(stderr,"Setting RTL resolution with ^[b*t R\n");
  }
  else{
    fprintf(out,"\x1b*t%dH",image->width*720/dpi_x); /* width (decipoints) */
    fprintf(out,"\x1b*t%dV",image->height*720/dpi_y); /* raster height ( " ) */
    if (debug>1) fprintf(stderr,"Setting RTL resolution with ^[b*t H and V\n");
  } 
  
  if (image->depth==24){
    if (debug) fprintf(stderr,"Writing 24 bit RGB RTL image %s\n",
                       (image->compression==RLE24)?
                       "with packbits compression":"");
    fprintf(out,"\x1b*v6W");  /* Can't have nulls in string.. */
    fputc(0,out);
    fputc(3,out);
    fputc(3,out);
    fputc(8,out);
    fputc(8,out);
    fputc(8,out);
  }
  else if (image->colour==PALETTE){
    if (debug) fprintf(stderr,"Writing paletted RTL image %s\n",
                       (image->compression==RLE24)?
                       "with packbits compression":"");
    fprintf(out,"\x1b*v6W");  /* Can't have nulls in string.. */
    fputc(0,out);
    fputc(1,out);
    fputc((char)image->depth,out);
    fputc(8,out);
    fputc(8,out);
    fputc(8,out);
    /* Set up palette */
    for(i=0;i<image->pal_len;i++){
      fprintf(out,"\x1b*v%da%db%dc%dI",image->palette[3*i],
              image->palette[3*i+1],image->palette[3*i+2],i);
    }
  }
  else if (image->colour==GREY){ /* Best treated as paletted,
                                    after constructing
                                    greyscale palette?! */
    if (debug) fprintf(stderr,"Writing greyscale RTL image %s\n",
                       (image->compression==RLE24)?
                       "with packbits compression":"");
    fprintf(out,"\x1b*v6W");  /* Can't have nulls in string.. */
    fputc(0,out);
    fputc(1,out);
    fputc((char)image->depth,out);
    fputc(8,out);
    fputc(8,out);
    fputc(8,out);
    /* Set up palette */
    j=0;
    if (image->depth==8)
      grey_step=1;
    else if (image->depth==4)
      grey_step=17;
    else if (image->depth==2)
      grey_step=85;
    else{
      fprintf(stderr,"Error: grey image depth of %d in write_rtl\n",
              image->depth);
      exit(1);
    }
    for(i=0;i<(1<<image->depth);i++){
        fprintf(out,"\x1b*v%da%db%dc%dI",j,j,j,i);
        j+=grey_step;
    }
  }
  else if (image->depth!=1){
    fprintf(stderr,"Error: unsupported image in write_rtl\n");
    exit(1);
  }
  
  fprintf(out,"\x1b*r3A"); /* Start scaled raster graphics */

  p=image->data;

  if (image->depth==1){
    if (debug) fprintf(stderr,"Writing 1 bit per pixel RTL image\n");
    row_len=(image->width+7)/8;
    for (i=0;i<image->height;i++){
      fprintf(out,"\x1b*b0M"); /* uncompressed row */
      fprintf(out,"\x1b*b%dW",row_len); /* data */
      /* we need to invert 1s and 0s */
      for(j=0;j<row_len;j++) p[j]=~p[j];
      fwrite(p,row_len,1,out);          /* data */
      p+=row_len;
    }
  }
  else{
    row_len=(image->width*image->depth+7)/8;
    rc=malloc(2*row_len+1);
    rs=malloc(2*row_len+1);
#ifdef DEBUG
    check=malloc(2*row_len+1);
#endif
    if ((!rc)||(!rs)){
      fprintf(stderr,"Malloc error in write_rtl\n");
      exit(1);
    }
    written=0;
    if (image->compression==RLE24){
      c=to_rle(p,rc,row_len)-1;  /* HP's RTL has no EOD byte, hence -1 */
      if (c<0.9*row_len){
        fprintf(out,"\x1b*b2M"); /* compressed row */
        fprintf(out,"\x1b*b%dW",c);
        fwrite(rc,c,1,out);
        pack++;
        written=1;
      }
    }
    if (!written){
      fprintf(out,"\x1b*b0M"); /* uncompressed row */
      fprintf(out,"\x1b*b%dW",row_len); /* data */
      fwrite(p,row_len,1,out);          /* data */
    }
    p+=row_len;
    for (i=1;i<image->height;i++){
      if (memcmp(p,p-row_len,row_len)){
        written=0;
        if (image->compression==RLE24){
          c=to_rle(p,rc,row_len)-1;
          s=seedrow(p,p-row_len,rs,row_len);
#if DEBUG
          unseedrow(rs,p-row_len,check,s,row_len);
          for(j=0;j<row_len;j++){
            if (check[j]!=p[j]){
              fprintf(stderr,"Decompress fails row %d byte %d\n",i,j);
              (*(int*)NULL)++;
            }
          }
#endif
          if ((s<c)&&(s<0.9*row_len)){
            fprintf(out,"\x1b*b3M"); /* compressed row */
            fprintf(out,"\x1b*b%dW",s);
            fwrite(rs,s,1,out);
            seed++;
            written=1;
          }
          else if (c<0.9*row_len){
            fprintf(out,"\x1b*b2M"); /* compressed row */
            fprintf(out,"\x1b*b%dW",c);
            fwrite(rc,c,1,out);
            pack++;
            written=1;
          }
        }
        if (!written){
          fprintf(out,"\x1b*b0m%dW",row_len); /* uncompressed row with data */
          fwrite(p,row_len,1,out);            /* data */
        }
        p+=row_len;
      }
      else{
        fprintf(out,"\x1b*b3m0W"); /* repeat previous row */
        if (debug>3) fprintf(stderr,"Row %d is duplicate\n",i);
        p+=row_len;
        dup++;
      }
    }
    free(rs);
    free(rc);
  }

  fprintf(out,"\x1b*rC"); /* end raster graphics */
  fprintf(out,"\x1b%%0B");  /* enter HPGL */
  fprintf(out,"PG;");   /* End and print current page */
  fprintf(out,"\033E");    /* reset */
  fprintf(out,"\x1b%%-12345X");
  fprintf(out,"@PJL EOJ NAME=\"RTL\"\r\n");

  if (debug){
    fprintf(stderr,"%d of %d rows duplicates (%.1f%%)\n",
	    dup,image->height,(100.0*dup)/image->height);
    if (pack)
      fprintf(stderr,"%d of %d rows packbits (%.1f%%)\n",
	      pack,image->height,(100.0*pack)/image->height);
    if (seed)
      fprintf(stderr,"%d of %d rows seedrow (%.1f%%)\n",
	      seed,image->height,(100.0*seed)/image->height);
    fprintf(stderr,"%d of %d rows uncompressed (%.1f%%)\n",
            image->height-dup-pack-seed,image->height,
            (100.0*(image->height-dup-pack-seed))/image->height);
  }
}

static unsigned int seedrow(unsigned char *row, unsigned char *last,
                            unsigned char *out, unsigned int row_len){
  unsigned char ctrl,*ptr,*p;
  unsigned int off,change,i,j;

  p=out;
  i=0;
  
  while(i<row_len){
    off=0;
    while((i<row_len)&&(row[i]==last[i])) {i++;off++;}
    if (i==row_len) break;
    ptr=row+i;
    change=0;
    i++;
    while((i<row_len)&&(row[i]!=last[i])&&(change<7)){
      i++;
      change++;
    }
    /* Write control byte(s) */
    ctrl=change<<5;
    if (off<=30){
      ctrl+=off;
      *(p++)=ctrl;
    }
    else{
      ctrl+=31;
      off-=31;
      *(p++)=ctrl;
      while(off>=255){
        *(p++)=255;
        off-=255;
      }
      *(p++)=(unsigned char)off;
    }
    /* Write replacement byte(s) */
    for(j=0;j<=change;j++)
      *(p++)=*(ptr++);
  }

  return(p-out);
}

#if DEBUG
static unsigned int unseedrow(unsigned char *in, unsigned char *last,
                              unsigned char *out, unsigned int data_len,
                              unsigned int row_len){
  unsigned char ctrl,*ptr,*p;
  unsigned int off,change,i,j;

  i=j=0;
  p=out;

  while(i<data_len){
    ctrl=in[i++];
    change=((unsigned int)ctrl>>5)+1;
    off=ctrl&31;
    if (off==31){
      while (in[i]==255){
        off+=255;
        i++;
      }
      off+=in[i++];
    }
    for(j=0;j<off;j++)
      *(p++)=*(last++);
    for(j=0;j<change;j++){
      *(p++)=in[i++];
      last++;
    }
  }
  while(p<(out+row_len)) *(p++)=*(last++);
}
#endif
