/* Copyright (c) 2017 MJ Rutter 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street,
 * Fifth Floor, Boston, MA  02110-1301, USA.
 */ 

#include <stdio.h>
#include <stdlib.h>

#include "bmp2eps.h"

void B64p(FILE* fout, unsigned char* in, int len, int line_len);
void write_png(FILE *out, struct bitmap *image);

void write_svg(FILE *out, struct bitmap *image, int type){
  FILE *tmp;
  long size;
  unsigned char *buff;
  int debug;

  debug=image->flags&DEBUG_MASK;
  
  if (type==0){
    fprintf(out,"<svg width=\"%d\" height=\"%d\"\n",
	    image->width,image->height);
    fprintf(out,"xmlns=\"http://www.w3.org/2000/svg\" xmlns:xlink=\"http://www.w3.org/1999/xlink\"\n>\n");
    fprintf(out,"<image width=\"%d\" height=\"%d\"\n xlink:href=\"data:image/",
	    image->width,image->height);
  }
  else{
    fprintf(out,"<img width=\"%d\" height=\"%d\" alt=\"\" src=\"data:image/",
	    image->width,image->height);
  }

  if (image->compression==DCT){
    fprintf(out,"jpeg;base64,");
    if (type) fprintf(out,"\n");
    B64p(out,image->data,image->data_len,type?76:0);
  }
  else{
    tmp=tmpfile();
    write_png(tmp,image);
    size=ftell(tmp);
    if (debug) fprintf(stderr,"Wrote PNG of length %d, now reading\n",
		       (int)size);
    buff=malloc(size);
    if (!buff){
      fprintf(stderr,"Malloc error for PNG file\n");
      exit(1);
    }
    fseek(tmp,0,SEEK_SET);
    if (fread(buff,1,size,tmp)!=size){
      fprintf(stderr,"Error reading back PNG file\n");
      exit(1);
    }
    fclose(tmp);
    fprintf(out,"png;base64,");
    if (type) fprintf(out,"\n");
    B64p(out,buff,(int)size,type?76:0);
    free(buff);
  }
    
    
  fprintf(out,"\"/>\n");
  if(type==0) fprintf(out,"</svg>\n");

}
